HardcoreHandler = class(Turbine.UI.Control)
function HardcoreHandler:Constructor()
    Turbine.UI.Control.Constructor(self)

    self.characterData = PrimeUtils.Utils.Handlers.LoadTable(Turbine.DataScope.Character, "Hardcore")

    self.characterOptions = PrimeUtils.Utils.Handlers.LoadTable(Turbine.DataScope.Character, "HardcoreOptions") or self:GetBaseOptions()

    self.data = PrimeUtils.Utils.Handlers.LoadTable(Turbine.DataScope.Character, "HardcoreData") or {
        Difficulty = 0,
    }

    --self.characterData = nil
    --self.characterData.Accepted_Quests = {}
    --self.characterData.Accepted_Deeds = {}
    --self.characterData.Completed_Quests = {}
    --self.characterData.Completed_Deeds = {}

    self.player = Turbine.Gameplay.LocalPlayer.GetInstance()
    self.playerName = self.player:GetName()
    self.equipment = self.player:GetEquipment()
    self.party = self.player:GetParty()

    self.startTime = Turbine.Engine.GetGameTime()

    self.button = WindowButton(self, self.characterOptions)
    
    self.achievementHandlers = {}

    self.dayOfTheWeekTable = {
        [1] = "Sunday",
        [2] = "Monday",
        [3] = "Tuesday",
        [4] = "Wednesday",
        [5] = "Thursday",
        [6] = "Friday",
        [7] = "Saturday",
    }

    self.difficultyTable = {
        ["Difficulty 0: Normal"] = 0,
        ["Difficulty 1: Adventurous"] = 1,
        ["Difficulty 2: Daring"] = 2,
        ["Difficulty 3: Fearless"] = 3,
        ["Difficulty 4: Fearless +1"] = 4,
        ["Difficulty 5: Fearless +2"] = 5,
        ["Difficulty 6: Fearless +3"] = 6,
        ["Difficulty 7: Heroic"] = 7,
        ["Difficulty 8: Heroic +1"] = 8,
        ["Difficulty 9: Heroic +2"] = 9,
    }

    self.missionTable = {}

    for key, value in pairs(InstanceController:GetDelvingTable()) do
        self.missionTable[value] = true
    end

    self.dungeonNames = InstanceController:GetDungeonNameTable()

    self.dungeonTable = {
        -- Annuminas
        ["Glinghant Corrupted"] = self.dungeonNames[46],
        ["New Quest: Keep It from the Enemy"] = self.dungeonNames[46],
        ["End Dolvaethor's Menace"] = self.dungeonNames[47],
        ["New Home for the Relics of Old"] = self.dungeonNames[47],
        ["Balhest, Lord of Ost Elendil"] = self.dungeonNames[48],
        ["Blood-stained Runes"] = self.dungeonNames[48],

        -- Fornost
        ["Fornost: Rhavameldir, the Wraith of Earth"] = self.dungeonNames[79],
        ["Fornost: The Jern-elyg"] = self.dungeonNames[79],
        ["Fornost: Weapons of War"] = self.dungeonNames[79],
        ["Fornost: Brogadan, Wraith of Fire"] = self.dungeonNames[80],
        ["Fornost: Leaders of Orcs"] = self.dungeonNames[80],
        ["Fornost: Idols of Sauron"] = self.dungeonNames[80],
        ["Fornost: Megoriath, the Wraith of Water"] = self.dungeonNames[81],
        ["Fornost: The Forgotten Company"] = self.dungeonNames[81],
        ["Fornost: The Brack"] = self.dungeonNames[81],
        ["Fornost: Remmenaeg, Wraith of Shadow"] = self.dungeonNames[82],
        ["Fornost: Ambassadors from Angmar"] = self.dungeonNames[82],
        ["Fornost: Free the Fallen"] = self.dungeonNames[82],

        -- Barrows
        ["Sambrog"] = self.dungeonNames[49],
        ["Accursed Fell Spirit"] = self.dungeonNames[49],
        ["Thadúr the Ravager"] = self.dungeonNames[50],
        ["Disturbed Remains"] = self.dungeonNames[50],
        ["Gaerdring and Gaerthel"] = self.dungeonNames[51],
        ["Ancient Barrow-wights"] = self.dungeonNames[51],

        -- Garth agarwen
        ["Challenge: Grimbark"] = self.dungeonNames[27],
        ["Optional Objective: Darkwater"] = self.dungeonNames[27],
        ["Optional Objective: Key Locations"] = self.dungeonNames[27],
        ["Optional Objective: Bog-lurkers"] = self.dungeonNames[27],
        --["Optional Objective: Wights"] = self.dungeonNames[27],
        ["Challenge: Ivar"] = self.dungeonNames[28],
        ["Optional Objective: Decayed"] = self.dungeonNames[28],
        ["Challenge: Naruhel"] = self.dungeonNames[29],
        ["Challenge: Dúnlang"] = self.dungeonNames[29],
        --["Optional Objective: Wights"] = self.dungeonNames[29],
    }

    self.dungeonBossTable = {
        -- Annuminas
        ["Nengon"]  = self.dungeonNames[46],
        ["Dolvaethor"] = self.dungeonNames[47],
        ["Balhest"] = self.dungeonNames[48],

        -- Fornost
        ["Rhavameldir"] = self.dungeonNames[79],
        ["Brogadan"] = self.dungeonNames[80],
        ["Megoriath"] = self.dungeonNames[81],
        ["Remmenaeg"] = self.dungeonNames[82],

        -- Barrows
        ["Sambrog"] = self.dungeonNames[49],
        ["Thadúr the Ravager"] = self.dungeonNames[50],
        ["Gaerdring"] = self.dungeonNames[51],
        ["Gaerthel"] = self.dungeonNames[51],

        -- Garth agarwen
        ["Grimbark"] = self.dungeonNames[27],
        ["Ivar the Blood-hand"] = self.dungeonNames[28],
        ["Naruhel"] = self.dungeonNames[29],
    }

    self.experienceItems = {
        ["Outrider's Token"] = true,
        ["Derudh's Stone"] = true,
        ["Fateful Gundabad Earring"] = true,
        ["Earring of the Dead City"] = true,
        ["Weightless Earring of the High Elves"] = true,
        ["Token of the Hornburg"] = true,
        ["Shield of the Hammerhand"] = true,
    }

    self.experienceBuffs = {
        ["Tome of Extraordinary Experience"] = true,
    }

    self.permittedDungeons = {
        [self.dungeonNames[46]] = true,
        [self.dungeonNames[47]] = true,
        [self.dungeonNames[48]] = true,
        [self.dungeonNames[79]] = true,
        [self.dungeonNames[80]] = true,
        [self.dungeonNames[81]] = true,
        [self.dungeonNames[82]] = true,
        [self.dungeonNames[49]] = true,
        [self.dungeonNames[50]] = true,
        [self.dungeonNames[51]] = true,
        [self.dungeonNames[27]] = true,
        [self.dungeonNames[28]] = true,
        [self.dungeonNames[29]] = true,
    }



    self.levelAchievements = {
        [10] = 31,
        [20] = 32,
        [30] = 33,
        [40] = 34,
        [50] = 35,
    }

    self.specialInstanceQuests = {
        ["Instance: Oakheart's Might"] = true,
    }

    self.achievements = Hardcore.Hardcore.Achievements.GetAchievements()

    self.introWindow = IntroWindow(self, self.characterData ~= nil)

    if self.characterData then
        self.hardcoreParty = PrimeUtils.Utils.Tablelength(self.characterData.Party) > 0
        if not(self.characterData.Fail) then
            self:SetBaseHardcoreCallbacks()

            self:SetTrackingCallbacks()

            -- Load the achievements
            self:LoadAchievementTracking()

            -- Display monitoring
            self.monitorPopup = MonitorPopup()

            local partyCheck

            if PrimeUtils.Utils.Tablelength(self.characterData.Party) > 0 then
                -- If you have a party

                if self.party then
                    local checkCount = 0
                    local partyCount = self.party:GetMemberCount()
                    for i = 1, partyCount do
                        local name = self.party:GetMember(i):GetName()
                        if self.characterData.Party[name] then
                            checkCount = checkCount + 1
                        end
                    end

                    if checkCount ~= PrimeUtils.Utils.Tablelength(self.characterData.Party) then
                        self:PartyIncomplete()
                    else    
                        partyCheck = true
                    end
                else
                    self:PartyIncomplete()
                end
            end

            if self.party then
                if self.data.Location ~= "Bree-town" then
                    if PrimeUtils.Utils.Tablelength(self.characterData.Party) > 0 and 
                    PrimeUtils.Utils.Tablelength(self.characterData.Party) ~= (self.party:GetMemberCount() - 1) then
                        self:IllegalParty()
                    elseif partyCheck == nil then
                        self:IllegalParty()
                    end
                end
            end
        else
            self.gameOverWindow = GameOverWindow(self.characterData)
        end
    end
end

function HardcoreHandler:PartyIncomplete()
    self.partyIncomplete = true
    self.timer = 600
    self.partyUpdateCallback = AddCallback(Updater, "ExecuteUpdate", function(delta)
        self.timer = self.timer - delta

        if self.timer < 0 and self.partyIncomplete then
            self:Fail("Party Incomplete", "Failed to form the party", "")
            RemoveCallback(Updater, "ExecuteUpdate", self.partyUpdateCallback)
            self.partyUpdateCallback = nil
        end
    end)
    ChatOutput("<rgb=#FF0000>Gather your group within 10 minutes to avoid failing the hardcore challenge!")
end

function HardcoreHandler:IllegalParty()
    self.illegalParty = true
    self.illegalPartyTimer = 60
    self.illegalPartyUpdateCallback = AddCallback(Updater, "ExecuteUpdate", function(delta)
        self.illegalPartyTimer = self.illegalPartyTimer - delta

        if self.illegalPartyTimer < 0 and self.illegalParty then
            self:Fail("Illegal Party", "Party in an illegal location", "")
            RemoveCallback(Updater, "ExecuteUpdate", self.illegalPartyUpdateCallback)
            self.illegalPartyUpdateCallback = nil
        end
    end)
    ChatOutput("<rgb=#FF0000>You are only allowed to be grouped for dungeons within Bree-town, go to Bree or leave the group within 60 seconds to avoid failing the hardcore challenge!")
end

function HardcoreHandler:ShowWindow()
    self.introWindow:SetVisible(not(self.introWindow:IsVisible()))
end

function HardcoreHandler:GetBaseOptions()
    return {
        buttonX = 500,
        buttonY = Turbine.UI.Display.GetHeight() - 200,
    }
end

function HardcoreHandler:GetAchievementData()
    return self.achievements or {}
end

function HardcoreHandler:GetAchievements()
    if self.characterData and self.characterData.Achievements_Attempting then
        return self.characterData.Achievements_Attempting
    else 
        return {}
    end
end

function HardcoreHandler:GetAchievementsCompleted()
    if self.characterData and self.characterData.Achievements_Done then
        return self.characterData.Achievements_Done
    else 
        return {}
    end
end

function HardcoreHandler:Begin(achievementData, partyData)
    if self.player:GetLevel() ~= 1 then
        ChatOutput("You are not level 1 and cannot start the hardcore challenge")
        return 
    end

    self.characterData = self:GetBaseData()

    self.characterData.Achievements_Attempting = achievementData
    self.characterData.Party = partyData

    ReloadPlugin()
end

function HardcoreHandler:GetBaseData()
    return
    {
        Start = Turbine.Engine.GetDate(),
        Quest_Completions = {},
        Monster_Kills = {},
        Experience = 0,
        Dungeons = {},
        Level_Log = {},
        Achievements_Attempting = {},
        Achievements_Done = {},
        Achievements_Failed = {},
        Party = {},
        Fail = false,
        PlayTime = 0,
        Last_Level_PlayTime = 0,
        Accepted_Quests = {},
        Accepted_Deeds = {},
        Completed_Quests = {},
        Completed_Deeds = {},
        Fail_Reason = "No fail",
    }
end

function HardcoreHandler:SetBaseHardcoreCallbacks()

    self.CallbacksSet = true

    for i = 1, PlayerEffectsHandler:GetPlayerEffects():GetCount() do
        if self.experienceBuffs[PlayerEffectsHandler:GetPlayerEffects():Get(i):GetName()] then 
            self:Fail("Experience Buff", "Used an experience tome", 0x41153603)
        end
    end

    self.effectCallback = AddCallback(PlayerEffectsHandler, "EffectAdded", function(effect)
        if self.experienceBuffs[effect:GetName()] then
            self:Fail("Experience Buff", "Used an experience tome", 0x41153603)
        end
    end)

    for i = 1, self.equipment:GetSize() do
        if self.equipment:GetItem(i) then
            if string.find(self.equipment:GetItem(i):GetItemInfo():GetName(), "Incomparable Mathom-hunter's") then
                self:Fail("Hobbit Armour", "Equipped hobbit present armour", "")
            end

            if self.experienceItems[self.equipment:GetItem(i):GetItemInfo():GetName()] then
                self:Fail("Experience Item", "Equipped an experience item", "")
            end
        end
    end

    self.itemEquippedCallback = AddCallback(self.equipment, "ItemEquipped", function(sender, args)
        local item = self.equipment:GetItem(args.Index)
        if item ~= nil then 
            local itemInfo = item:GetItemInfo()
            if string.find(itemInfo:GetName(), "Incomparable Mathom-hunter's") then
                self:Fail("Hobbit Armour", "Equipped hobbit present armour", "")
            end

            if self.experienceItems[itemInfo:GetName()] then
                self:Fail("Experience Item", "Equipped an experience item", "")
            end
        end	
    end)

    self.locationCallbacks = AddCallback(LocationHandler, "Location", function(args)
        if args.Entered then
            if args.Location == "Ettenmoors" then
                self:Fail()
            end

            self.data.Location = args.Location

            if self.illegalParty and args.Location == "Bree-town" then
                if self.illegalPartyUpdateCallback then
                    RemoveCallback(Updater, "ExecuteUpdate", self.illegalPartyUpdateCallback)
                    self.illegalPartyUpdateCallback = nil
                    ChatOutput("<rgb=#00FF00>Bree-town reached!")
                end
            end
            
            if self.dungeonCompletedCallback then
                RemoveCallback(Updater, "ExecuteUpdate", self.dungeonCompletedCallback)
                self.dungeonCompletedCallback = nil
            end
        end
    end)

    self.standardChatCallback = AddCallback(ChatEventHandler, "Standard", function(message)

        -- Place bid
        if string.match(message, "Your bid for (.*) has been placed successfully.") then
            self:Fail("Austion House", "Used the Auction House", "")
        end

        -- Win acution
        if string.match(message, "You have won the auction for (.*).") then
            self:Fail("Austion House", "Used the Auction House", "")
        end

        -- Swift travel
        if string.match(message, "You purchased a ride to (.*) - Swift Travel for (.*).") then
            self:Fail("Swift Travel", "Used Swift Travel", "")
        end

        if not self.hardcoreParty then
            if string.match(message, "Trade finished.") then
                self:Fail("Traded", "Traded", "")
            end
        end

        if self.hardcoreParty then
            local tradeOffer = string.match(message, "(.*) wishes to trade with you.")
            if tradeOffer then
                self.tradeOffer = tradeOffer
            end

            if self.tradeOffer ~= "" and string.match(message, "Trade finished.") and not self.characterData.Party[self.tradeOffer] then
                self:Fail("Traded", "Traded with an outside member", "")
            end
        end

        if string.find(message, "You have been playing for:") then
            local timePlayed = PrimeUtils.Utils.CalculateTotalTimePlayed(message)

            local totalPlaytime =  self.characterData.PlayTime + (Turbine.Engine.GetGameTime() - self.startTime)

            self.characterData.VerificationPlayedTime = totalPlaytime/timePlayed*100

            self.characterData.VerificationCheckTime = Turbine.Engine.GetGameTime()

            self.introWindow:SetVerification(self.characterData.VerificationPlayedTime, self.characterData.VerificationCheckTime)
        end

        if string.match(message, "Purchased travel for %d+ Mithril Coins.") then
            self:Fail("Failed", "Used Mithril Travel", "")
        end

        if string.find(message, "You spent Mithril Coin to travel") then
            self:Fail("Failed", "Used Mithril Travel", "")
        end
    end)

    self.deathCallback = AddCallback(ChatEventHandler, "Defeated", function(args)
        if args.Self and args.Name == self.playerName then
            if args.Death_Source and args.Death_Source == "Misadventure" and self.deathImmune then return end
            self:Fail()
        end

        if self.hardcoreParty then
            if self.characterData.Party[args.Name] then
                for i = 1, self.party:GetMemberCount() do
                    local partyMember = self.party:GetMember(i)
                    if partyMember:GetName() == args.Name then
                        if partyMember:GetMorale() <= 0 then
                            self:Fail("Fellowship Failed", args.Name .. " of your fellowship died.")
                        end
                    end
                end
            end
        end
    end)

    self.newDeedCallback = AddCallback(ChatEventHandler, "NewDeed", function(args)
        self.characterData.Accepted_Deeds[args.Name] = true
    end)

    self.newQuestCallback = AddCallback(ChatEventHandler, "QuestAccepted", function(args)
        self.characterData.Accepted_Quests[args.Name] = true

        if self.specialInstanceQuests[args.Name] then
            self.deathImmune = true

            if not(self.specialLocationSwap) then
                self.specialLocationSwap = AddCallback(LocationHandler, "Location", function(args)
                    if args.Entered then
                        self.deathImmune = false
                        RemoveCallback(LocationHandler, "Location", self.specialLocationSwap)
                        self.specialLocationSwap = nil
                    end
                end)    
            end
        end

        if self.dungeonTable[args.Name] then
            if self.characterData.Dungeons[self.dungeonTable[args.Name]] then
                self:DungeonHandling()
            end
        end
    end)

    self.doneQuestOrDeedCallback = AddCallback(ChatEventHandler, "QuestOrDeedCompleted", function(args)
        if self.characterData.Accepted_Quests[args.Name] then
            if not(self.characterData.Completed_Quests[args.Name]) then
                self.characterData.Completed_Quests[args.Name] = 0
            end
            self.characterData.Completed_Quests[args.Name] = self.characterData.Completed_Quests[args.Name] + 1
        end

        if self.characterData.Accepted_Deeds[args.Name] then
            if not(self.characterData.Completed_Deeds[args.Name]) then
                self.characterData.Completed_Deeds[args.Name] = 0
            end
            self.characterData.Completed_Deeds[args.Name] = self.characterData.Completed_Deeds[args.Name] + 1
        end

        if self.difficultyTable[args.Name] then
            local newDifficulty = self.difficultyTable[args.Name]
            if newDifficulty > self.data.Difficulty and self.player:GetLevel() > 10 then return end
            self.data.Difficulty = newDifficulty
        end

        if self.dungeonTable[args.Name] then
            self:DungeonAchievement(self.dungeonTable[args.Name])
            self.characterData.Dungeons[self.dungeonTable[args.Name]] = true
        end

        if self.missionTable[args.Name] then
            self:Fail("Mission Fail", "Completed a mission", "")
        end
        
        if self.player:GetLevel() <= 45 and args.Name == "Bilbo's Buttons"  then
            self:AchievementAchieved(55)
        end

        if self.player:GetLevel() <= 25 and args.Name == "Executioner of the Wicked"  then
            self:AchievementAchieved(56)
        end

        if args.Name == "VIP Rewards" then
            self:Fail("VIP Quest", "Completed the VIP quest", "")
        end
    end)

    self.instanceEnterCallback = AddCallback(InstanceController, "InstanceEvent", function(args)
        if args.Instance_State then
            if self.characterData.Dungeons[args.Instance_Name] then
                self:DungeonHandling()
            end
        else
            if self.dungeonCompletedCallback then
                RemoveCallback(Updater, "ExecuteUpdate", self.dungeonCompletedCallback)
                self.dungeonCompletedCallback = nil
            end
        end

        if not self.permittedDungeons[args.Instance_Name] then
            self:IllegalDungeonHandling()
        end
    end)

    self.killCallback = AddCallback(ChatEventHandler, "Defeated", function(args)
        if not args.Death_Source and self.dungeonBossTable[args.Name] and self.characterData.Dungeons[self.dungeonBossTable[args.Name]] then
            self:Fail("Duplicate dungeon", "Already completed the dungeon", "")
        end
    end)


    self.illegalSkills = {
        "Return Home",
        "Return to (.*)",
        --"Guide to (.*)",
        "Desperate Flight",
        "Return Home %d+",
        "Still As Death",
        "Hobbit-silence" -- hobbit one?
    }

    self.skillCallbacks = {}

    for key, value in pairs(SkillEventHandler:GetSkills()) do
        for _, pattern in pairs(self.illegalSkills) do
            if string.match(key, pattern) then
                table.insert(self.skillCallbacks, AddCallback(SkillEventHandler, key, function(skill) 
                    self:Fail("Illegal Skill", "Used an Illegal skill", "")
                end))
            end
        end

    end
end

function HardcoreHandler:DungeonHandling()
    if not self.dungeonCompletedCallback then
        self.dungeonTimer = 60
        ChatOutput("<rgb=#FF0000>You have already completed this dungeon, leave within 60 seconds to avoid failing the hardcore challenge!")
        self.dungeonCompletedCallback = AddCallback(Updater, "ExecuteUpdate", function(delta)
            self.dungeonTimer = self.dungeonTimer - delta

            if self.dungeonTimer < 0 then
                self:Fail("Duplicate dungeon", "Already completed the dungeon", "")
                RemoveCallback(Updater, "ExecuteUpdate", self.dungeonCompletedCallback)
                self.dungeonCompletedCallback = nil
            end
        end)
    end
end

function HardcoreHandler:IllegalDungeonHandling()
    if not self.dungeonCompletedCallback then
        self.dungeonTimer = 60
        ChatOutput("<rgb=#FF0000>This dungeon is not allowed to be completed, leave in 60 seconds or fail the hardcore challenge")
        self.dungeonCompletedCallback = AddCallback(Updater, "ExecuteUpdate", function(delta)
            self.dungeonTimer = self.dungeonTimer - delta

            if self.dungeonTimer < 0 then
                self:Fail("Illegal dungeon", "Illegal dungeon", "")
                RemoveCallback(Updater, "ExecuteUpdate", self.dungeonCompletedCallback)
                self.dungeonCompletedCallback = nil
            end
        end)
    end
end

function HardcoreHandler:LoadAchievementTracking()
    local achievementTable = {
        ["I Can See You"] = Achievements.AchievementICanSeeYou,
        ["Runestone Mastery"] = Achievements.AchievementRuneStoneMastery,
        ["Always Surrender"] = Achievements.AchievementAlwaysSurrender,
        ["No Way Out"] = Achievements.AchievementNoWayOut,
        ["We will Fall This Day"] = Achievements.AchievementWeWillFallThisDay,
        ["Naturalist"] = Achievements.AchievementNaturalist,
        ["Popped"] = Achievements.AchievementPopped,
        ["Perfectly Balanced"] = Achievements.AchievementPerfectlyBalanced,
        ["Pyromaster"] = Achievements.AchievementPyromaster,
        ["Loose Hands"] = Achievements.AchievementLooseHands,
        ["Tied Hands"] = Achievements.AchievementTiedHands,
        ["Back To Basics"] = Achievements.AchievementBackToBasics,
        ["Frenzy Waiting Room"] = Achievements.AchievementFrenzyWaitingRoom,
        ["Fire, Light & Frost"] = Achievements.AchievementFireLightFrost,
        ["The Lone Wizard"] = Achievements.AchievementTheLoneWizard,
        ["The Lone Knight"] = Achievements.AchievementTheLonKnight,
        ["No health Potions"] = Achievements.AchievementNoHealhtPotions,
        ["The White Knight"] = Achievements.AchievementTheWhiteKnight,
        ["Nobody Got Time for That"] = Achievements.AchievementNobodyGotTimeForThat,
        ["Grounded"] = Achievements.AchievementGrounded,
        ["Blood Bath"] = Achievements.AchievementBloodBath,
        ["Design Hell"] = Achievements.AchievementDesignHell,
        ["Solitary Struggle"] = Achievements.AchievementSolitaryStruggle,
        ["All Gas No Breaks"] = Achievements.AchievementAllGasNoBreaks,
        ["Focused"] = Achievements.AchievementFocused,
        ["Lost The Path"] = Achievements.AchievementLostThePath,
        ["Nudist"] = Achievements.AchievementNudist,
        ["Not So Talented"] = Achievements.AchievementNotSoTalented,
        ["Self Made"] = Achievements.AchievementSelfMade,
    }


    for key, value in pairs(self.characterData.Achievements_Attempting) do
        if achievementTable[key] then 
            self.achievementHandlers[key] = achievementTable[key](self)
        end
    end

    self.achievementHandlers["Dungeons"] = Achievements.AchievementDungeons(self)
    self.achievementHandlers["Discovery"] = Achievements.AchievementDiscovery(self)
end

function HardcoreHandler:Fail(title, reason, icon)
    if true then
        --ChatOutput("Fail")
        --AchievementsPanel:AddItem(icon, "FAILED", title or "Failed", reason or "Failed", nil)
        --return
    end
    if self.characterData then
        self.characterData.Fail_Reason = reason
        self.characterData.Fail = true
        self.gameOverWindow = GameOverWindow(self.characterData)
        PrimeUtils.Utils.Handlers.SaveTable(Turbine.DataScope.Character, "HardcoreData", self.data)
        PrimeUtils.Utils.Handlers.SaveTable(Turbine.DataScope.Character, "Hardcore", self.characterData)
        PrimeUtils.Utils.Handlers.SaveTable(Turbine.DataScope.Character, "HardcoreBackup", self.characterData)
        PrimeUtils.Utils.Handlers.SaveTable(Turbine.DataScope.Account, "Hardcore" .. self.playerName, self.characterData)
        PrimeUtils.Utils.Handlers.SaveTable(Turbine.DataScope.Account, "HardcoreBackup" ..  self.playerName, self.characterData)
    end
    self:UnloadCallbacks()
end

function HardcoreHandler:UnloadCallbacks()
    if self.CallbacksSet then
        RemoveCallback(LocationHandler, "Location", self.locationCallbacks)
        RemoveCallback(ChatEventHandler, "Standard", self.standardChatCallback)
        RemoveCallback(ChatEventHandler, "Defeated", self.deathCallback)
        RemoveCallback(ChatEventHandler, "NewDeed", self.newDeedCallback)
        RemoveCallback(ChatEventHandler, "Defeated", self.killCallback)
        RemoveCallback(ChatEventHandler, "QuestAccepted", self.newQuestCallback)
        RemoveCallback(ChatEventHandler, "QuestOrDeedCompleted", self.doneQuestOrDeedCallback)
        RemoveCallback(PlayerEffectsHandler, "EffectAdded", self.effectCallback)
        RemoveCallback(ChatEventHandler, "PartyChange", self.partyChangeCallback)
        --RemoveCallback(self.equipment, "ItemEquipped", self.itemEquippedCallback)
        for key, value in pairs(self.skillCallbacks) do
            RemoveCallback(SkillEventHandler, key, value) 
        end
        if self.dungeonCompletedCallback then
            RemoveCallback(Updater, "ExecuteUpdate", self.dungeonCompletedCallback)
            self.dungeonCompletedCallback = nil
        end
    end
end

function HardcoreHandler:Unload()
    self:UnloadCallbacks()
    for key, value in pairs(self.achievementHandlers) do
        if value.Unload then
            value:Unload()
        end
    end

    if self.characterData then
        self:UpdateStats()

        self.characterOptions.buttonX, self.characterOptions.buttonY = self.button:GetPosition()

        PrimeUtils.Utils.Handlers.SaveTable(Turbine.DataScope.Character, "HardcoreOptions", self.characterOptions)
        PrimeUtils.Utils.Handlers.SaveTable(Turbine.DataScope.Character, "HardcoreData", self.data)
        PrimeUtils.Utils.Handlers.SaveTable(Turbine.DataScope.Character, "Hardcore", self.characterData)
        PrimeUtils.Utils.Handlers.SaveTable(Turbine.DataScope.Character, "HardcoreBackup", self.characterData)
        PrimeUtils.Utils.Handlers.SaveTable(Turbine.DataScope.Account, "Hardcore" .. self.playerName, self.characterData)
        PrimeUtils.Utils.Handlers.SaveTable(Turbine.DataScope.Account, "HardcoreBackup" ..  self.playerName, self.characterData)
    end
end

function HardcoreHandler:UpdateStats()
    if not(self.characterData.Fail) then
        self.characterData.PlayTime = self.characterData.PlayTime + (Turbine.Engine.GetGameTime() - self.startTime)
    end
end

function HardcoreHandler:GetCharacterData()
    return self.characterData
end

function HardcoreHandler:SetTrackingCallbacks()
    self.levelCallback = AddCallback(PlayerHandler, "LevelChanged", function(args)
        local totalPlaytime =  self.characterData.PlayTime + (Turbine.Engine.GetGameTime() - self.startTime)

        local date = Turbine.Engine.GetDate()
        table.insert(self.characterData.Level_Log, string.format("You reached level %s on %s %s/%s %02d:%02d:%02d after playing for %s on this level and %s in total.", args.Player_Level, self.dayOfTheWeekTable[date.DayOfWeek] or "??", date.Day, date.Month, date.Hour, date.Minute, date.Second, PrimeUtils.Utils.FormatTimeStringWritten(totalPlaytime - self.characterData.Last_Level_PlayTime), PrimeUtils.Utils.FormatTimeStringWritten(totalPlaytime)))

        ChatOutput(string.format("<rgb=#f79000>Level increased to Level %d, time taken for this level is %s.", args.Player_Level, PrimeUtils.Utils.FormatTimeString(totalPlaytime - self.characterData.Last_Level_PlayTime)))

        self.characterData.Last_Level_PlayTime = totalPlaytime

        if self.levelAchievements[args.Player_Level] then
            local achievementID = self.levelAchievements[args.Player_Level]

            self:AchievementAchieved(achievementID)
        end

        if args.Player_Level == 50 then
            self:Finish()
        end
    end)

    self.effectCallback = AddCallback(PlayerEffectsHandler, "EffectAdded", function(effect)
        local name = effect:GetName()
        if name == "Falling Injuries" or name == "Falling Injury Immunity" then
            self.characterData.Achievements_Failed["Bouncy"] = {["PlayTime"] = self.characterData.PlayTime + (Turbine.Engine.GetGameTime() - self.startTime), ["Level"] = self.player:GetLevel()}
        end
    end)

    self.partyChangeCallback = AddCallback(ChatEventHandler, "PartyChange", function(args)
        ReloadPlugin()
    end)
end

function HardcoreHandler:GetAchievementHint(achievement)
    if self.achievementHandlers[achievement] and self.achievementHandlers[achievement].GetHint then
        self.achievementHandlers[achievement]:GetHint()
    end
end

function HardcoreHandler:Finish()
    for key, value in pairs(self.characterData.Achievements_Attempting) do
        if not self.characterData.Achievements_Failed[key] then
            self:AchievementAchieved(value)
        end
    end

    if not self.characterData.Achievements_Failed["Bouncy"] then self:AchievementAchieved(30) end

    self.characterData.Completed50 = true
    self.completionWindow = CompletionWindow(self)
end

function HardcoreHandler:DungeonAchievement(dungeonName)
    self.achievementHandlers["Dungeons"]:HandleCompletion(dungeonName, self.characterData)
end

function HardcoreHandler:NameToID(name)
    for key, value in pairs(self.achievements) do
        if value.Title == name then return key end
    end
end

function HardcoreHandler:AchievementAchieved(id)

    if type(id) ~= "number" then
        ChatOutput(id)
        return
    end
    if self.characterData.Achievements_Done[id] then return end

    local totalPlaytime =  self.characterData.PlayTime + (Turbine.Engine.GetGameTime() - self.startTime)


    self.characterData.Achievements_Done[id] = {["PlayTime"] = totalPlaytime, ["Level"] = self.player:GetLevel(), ["Difficulty"] = self.data.Difficulty}

    local achievementData = self.achievements[id]

    AchievementsPanel:AddItem(achievementData.Icon, "Achievement Accomplished", achievementData.Title, achievementData.Description, "")
end